
ELEMENT_DOCUMENT:      str = 'PyutDocument'
TOP_LEVEL_ELEMENT:     str = 'PyutProject'
ELEMENT_GRAPHIC_CLASS: str = 'GraphicClass'
ELEMENT_GRAPHIC_LINK:  str = 'GraphicLink'
ELEMENT_MODEL_CLASS:   str = 'Class'
ELEMENT_MODEL_LINK:    str = 'Link'
ELEMENT_CONTROL_POINT: str = 'ControlPoint'

ELEMENT_MODEL_METHOD: str = 'Method'

ATTR_WIDTH:  str = 'width'
ATTR_HEIGHT: str = 'height'
ATTR_X:      str = 'x'
ATTR_Y:      str = 'y'
ATTR_NAME:   str = 'name'

ATTR_LINK_SOURCE_ANCHOR_X: str = 'srcX'
ATTR_LINK_SOURCE_ANCHOR_Y: str = 'srcY'

ATTR_LINK_DESTINATION_ANCHOR_X: str = 'dstX'
ATTR_LINK_DESTINATION_ANCHOR_Y: str = 'dstY'

ATTR_TYPE: str = 'type'

ATTR_SHOW_METHODS:      str = 'showMethods'
ATTR_SHOW_FIELDS:       str = 'showFields'
ATTR_SHOW_STEREOTYPE:   str = 'showStereotype'

ATTR_DISPLAY_PARAMETERS: str = 'displayParameters'

ELEMENT_MODEL_PARAM:       str = 'Param'
ELEMENT_MODEL_RETURN_TYPE: str = 'Return'
ATTR_DEFAULT_VALUE:  str = 'defaultValue'
ATTR_VISIBILITY:     str = 'visibility'

ELEMENT_PROJECT_V11:          str = 'PyutProject'
PROJECT_TYPE_V11:             str = 'type'
ELEMENT_DOCUMENT_V11:         str = 'PyutDocument'
ELEMENT_GRAPHIC_CLASS_V11:    str = 'OglClass'
ELEMENT_GRAPHIC_LINK_V11:     str = 'OglLink'
ELEMENT_CONTROL_POINT_V11:    str = 'ControlPoint'
ELEMENT_GRAPHIC_NOTE_V11:     str = 'OglNote'
ELEMENT_GRAPHIC_LOLLIPOP_V11: str = 'OglInterface2'

ELEMENT_LABEL_CENTER_V11:      str = 'LabelCenter'
ELEMENT_LABEL_SOURCE_V11:      str = 'LabelSource'
ELEMENT_LABEL_DESTINATION_V11: str = 'LabelDestination'

ATTR_WIDTH_V11:  str = 'width'
ATTR_HEIGHT_V11: str = 'height'
ATTR_X_V11:      str = 'x'
ATTR_Y_V11:      str = 'y'
ATTR_NAME_V11:   str = 'name'


ATTR_STEREOTYPE_V11:            str = 'stereotype'
ATTR_DISPLAY_METHODS_V11:       str = 'displayMethods'
ATTR_DISPLAY_PARAMETERS_V11:    str = 'displayParameters'
ATTR_DISPLAY_FIELDS_V11:        str = 'displayFields'
ATTR_DISPLAY_STEREOTYPE_V11:    str = 'displayStereotype'
ATTR_FILENAME_V11:              str = 'fileName'
ATTR_DESCRIPTION_V11:           str = 'description'

ELEMENT_MODEL_METHOD_V11: str = 'PyutMethod'
ATTR_VISIBILITY_V11:      str = 'visibility'
ATTR_RETURN_TYPE_V11:     str = 'returnType'

ELEMENT_MODEL_PARAMETER_V11: str  = 'PyutParameter'
ATTR_DEFAULT_VALUE_V11:      str  = 'defaultValue'
ATTR_TYPE_V11:               str  = 'type'

ELEMENT_MODEL_FIELD_V11:    str = 'PyutField'

ATTR_LINK_SOURCE_ANCHOR_X_V11: str = 'sourceAnchorX'
ATTR_LINK_SOURCE_ANCHOR_Y_V11: str = 'sourceAnchorY'

ATTR_LINK_DESTINATION_ANCHOR_X_V11: str = 'destinationAnchorX'
ATTR_LINK_DESTINATION_ANCHOR_Y_V11: str = 'destinationAnchorY'

ATTR_CARDINALITY_SOURCE_V11:      str = 'cardinalitySource'
ATTR_CARDINALITY_DESTINATION_V11: str = 'cardinalityDestination'

ATTR_CONTENT_V11:         str = 'content'
ATTR_ATTACHMENT_SIDE_V11: str = 'attachmentPoint'
